"""
HTML parsing.

Detect encoding from HTML header.
"""
from __future__ import unicode_literals
from .. import parsers
import re
import codecs
from ..filters import html_filter

RE_HTML_ENCODE = re.compile(
    br'''(?x)
    <meta(?!\s*(?:name|value)\s*=)(?:[^>]*?content\s*=[\s"']*)?(?:[^>]*?)[\s"';]*charset\s*=[\s"']*([^\s"'/>]*)
    '''
)


class HTMLDecoder(parsers.Decoder):
    """Detect HTML encoding."""

    def header_check(self, content):
        """Special HTML encoding check."""

        encode = None
        m = RE_HTML_ENCODE.search(content)
        if m:
            encode = m.group(1).decode('ascii')
        return encode


class HTMLParser(parsers.Parser):
    """Spelling Python."""

    FILE_PATTERNS = ('*.html', '*.htm')
    DECODER = HTMLDecoder

    def __init__(self, options, default_encoding='ascii'):
        """Initialization."""

        self.filter = html_filter.HTMLFilter(options)
        super(HTMLParser, self).__init__(options, default_encoding)

    def parse_file(self, source_file, encoding):
        """Parse HTML file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        content = [parsers.SourceText(self.filter.filter(text, encoding), source_file, encoding, 'html')]

        return content


def get_parser():
    """Return the parser."""

    return HTMLParser



"""Text parser."""
from __future__ import unicode_literals
from .. import parsers


class TextParser(parsers.Parser):
    """Text parser."""

    FILE_PATTERNS = ('*.txt', '*.text')


def get_parser():
    """Return the parser."""

    return TextParser



"""Main."""
from __future__ import unicode_literals
import sys
import argparse
from pyspelling import settings
from pyspelling import Aspell
from pyspelling import __version__


def run():
    """Main."""

    parser = argparse.ArgumentParser(prog='spellcheck', description='Spell checking tool.')
    # Flag arguments
    parser.add_argument('--version', action='version', version=('%(prog)s ' + __version__.version))
    parser.add_argument('--verbose', '-v', action='count', default=0, help="Verbosity level.")
    parser.add_argument('--name', '-n', action='store', default='', help="Specific spelling task by name to run.")
    parser.add_argument('--config', '-c', action='store', default='.spelling.yml', help="Spelling config.")
    args = parser.parse_args()

    fail = False
    config = settings.read_config(args.config)
    spelling = Aspell(config, name=args.name, verbose=args.verbose)
    for results in spelling.check():
        if results.error:
            fail = True
            print('ERROR: Could not read %s -- %s' % (results.context, results.error))
        elif results.words:
            fail = True
            print('Misspelled words:\n<%s> %s' % (results.category, results.context))
            print('-' * 80)
            for word in results.words:
                print(word)
            print('-' * 80)
            print('')
        elif args.verbose:
            print('CHECK: %s' % results.context)
    return fail


def main():
    """Main entry point."""

    sys.exit(run())


if __name__ == "__main__":
    main()



"""Spell check with Aspell or Hunspell."""
from __future__ import unicode_literals
import os
import fnmatch
import re
import codecs
import importlib
from . import util
from . import __version__

version = __version__.version
version_info = __version__.version_info


class Aspell(object):
    """Aspell spell check class."""

    DICTIONARY = 'dictionary.dic'

    def __init__(self, config, name='', verbose=0):
        """Initialize."""

        # General options
        self.name = name
        self.verbose = verbose
        self.dict_bin = os.path.abspath(self.DICTIONARY)
        self.documents = config.get('documents', [])
        self.dictionary = config.get('dictionary', [])

    def log(self, text, level):
        """Log level."""
        if self.verbose >= level:
            print(text)

    def normalize_utf(self, encoding):
        """Normalize UTF encoding."""

        if encoding == 'utf-8-sig':
            encoding = 'utf-8'
        if encoding.startswith('utf-16'):
            encoding = 'utf-16'
        elif encoding.startswith('utf-32'):
            encoding = 'utf-32'
        return encoding

    def setup_command(self, encoding, options, personal_dict):
        """Setup the command."""

        cmd = [
            'aspell',
            'list',
            '--encoding', codecs.lookup(encoding).name
        ]

        if personal_dict:
            cmd.extend(['--add-extra-dicts', personal_dict])

        allowed = {
            'conf-dir', 'data-dir', 'add-dict-alias', 'rem-dict-alias', 'dict-dir',
            'encoding', 'add-filter', 'rem-filter', 'add-filter-path', 'rem-filter-path',
            'mode', 'e', 'H', 't', 'n', 'add-extra-dicts', 'rem-extra-dicts', 'home-dir',
            'ingore', 'W', 'dont-ignore-case', 'ignore-case', 'lang', 'l', 'local-data-dir',
            'd', 'master', 'dont-normalize', 'normalize', 'dont-norm-required',
            'norm-required', 'norm-form', 'dont-norm-strict', 'norm-strict', 'per-conf',
            'p', 'personal', 'C', 'B', 'dont-run-together', 'run-together', 'run-together-limit',
            'run-together-min', 'use-other-dicts', 'dont-use-other-dicts', 'add-variety', 'rem-variety',
            'add-context-delimiters', 'rem-context-delimiters', 'dont-context-visible-first',
            'context-visible-first', 'add-email-quote', 'rem-email-quote', 'email-margin',
            'add-html-check', 'rem-html-check', 'add-html-skip', 'rem-html-skip', 'add-sgml-check',
            'rem-sgml-check', 'add-sgml-skip', 'rem-sgml-skip', 'dont-tex-check-comments',
            'tex-check-comments', 'add-tex-command', 'rem-tex-command', 'add-texinfo-ignore',
            'rem-texinfo-ignore', 'add-texinfo-ignore-env', 'rem-texinfo-ignore-env', 'filter'
        }

        for k, v in options.items():
            if k in allowed:
                key = ('-%s' if len(k) == 1 else '--%s') % k
                if isinstance(v, bool) and v is True:
                    cmd.append(key)
                elif isinstance(v, util.ustr):
                    cmd.extend([key, v])
                elif isinstance(v, int):
                    cmd.extend([key, util.ustr(v)])
                elif isinstance(v, list):
                    for value in v:
                        cmd.extend([key, util.ustr(value)])
        return cmd

    def check_spelling(self, sources, options, personal_dict):
        """Check spelling."""

        for source in sources:

            if source._has_error():
                yield util.Results([], source.context, source.category, source.error)
                continue

            text = source.text
            encoding = self.normalize_utf(source.encoding)
            if not source._is_bytes():
                for f, disallow in self.filters:
                    if source.category not in disallow:
                        text = f.filter(text, source.encoding)
                text = text.encode(encoding)
            self.log(text, 3)

            cmd = self.setup_command(encoding, options, personal_dict)
            self.log(str(cmd), 2)

            wordlist = util.console(cmd, input_text=text)
            yield util.Results([w for w in sorted(set(wordlist.split('\n'))) if w], source.context, source.category)

    def compile_dictionary(self, lang, wordlists, output):
        """Compile user dictionary."""

        output_location = os.path.dirname(output)
        if not os.path.exists(output_location):
            os.makedirs(output_location)
        if os.path.exists(output):
            os.remove(output)

        self.log("Compiling Dictionary...", 1)
        # Read word lists and create a unique set of words
        words = set()
        for wordlist in wordlists:
            with open(wordlist, 'rb') as src:
                for word in src.read().split(b'\n'):
                    words.add(word.replace(b'\r', b''))

        # Compile wordlist against language
        util.console(
            [
                'aspell',
                '--lang', lang,
                '--encoding=utf-8',
                'create',
                'master', output
            ],
            input_text=b'\n'.join(sorted(words)) + b'\n'
        )

    def skip_target(self, target):
        """Check if target should be skipped."""

        return self.skip_wild_card_target(target) or self.skip_regex_target(target)

    def skip_wild_card_target(self, target):
        """Check if target should be skipped via wildcard patterns."""

        exclude = False
        for pattern in self.excludes:
            if fnmatch.fnmatch(target, pattern):
                exclude = True
                break
        return exclude

    def skip_regex_target(self, target):
        """Check if target should be skipped via regex."""

        exclude = False
        for pattern in self.regex_excludes:
            if pattern.match(target, pattern):
                exclude = True
                break
        return exclude

    def is_valid_file(self, file_name, file_patterns):
        """Is file in current file patterns."""

        okay = False
        lowered = file_name.lower()
        for pattern in file_patterns:
            if fnmatch.fnmatch(lowered, pattern):
                okay = True
                break
        return okay

    def walk_src(self, targets, plugin):
        """Walk source and parse files."""

        # Override file_patterns if the user provides their own
        file_patterns = self.file_patterns if self.file_patterns else plugin.FILE_PATTERNS
        for target in targets:
            if not os.path.exists(target):
                continue
            if os.path.isdir(target):
                if self.skip_target(target):
                    continue
                for base, dirs, files in os.walk(target):
                    [dirs.remove(d) for d in dirs[:] if self.skip_target(os.path.join(base, d))]
                    for f in files:
                        file_path = os.path.join(base, f)
                        if self.skip_target(file_path):
                            continue
                        if self.is_valid_file(f, file_patterns):
                            yield plugin._parse(file_path)
            elif self.is_valid_file(target, file_patterns):
                if self.skip_target(target):
                    continue
                yield plugin._parse(target)

    def setup_spellchecker(self, documents):
        """Setup spell checker."""

        return documents.get('aspell', {})

    def setup_dictionary(self, documents):
        """Setup dictionary."""

        dictionary_options = documents.get('dictionary', {})
        output = os.path.abspath(dictionary_options.get('output', self.dict_bin))
        lang = dictionary_options.get('lang', 'en')
        wordlists = dictionary_options.get('wordlists', [])
        if lang and wordlists:
            self.compile_dictionary(lang, dictionary_options.get('wordlists', []), output)
        else:
            output = None
        return output

    def setup_excludes(self, documents):
        """Setup excludes."""

        # Read excludes
        self.excludes = documents.get('excludes', [])
        self.regex_excludes = [re.compile(exclude) for exclude in documents.get('regex_excludes', [])]

    def get_filters(self, documents):
        """Get filters."""
        self.filters = []
        for f in documents.get('filters', []):
            # Retrieve module and module options
            if isinstance(f, dict):
                name, options = list(f.items())[0]
            else:
                name = f
                options = {}
            if options is None:
                options = {}

            # Extract disallowed tokens
            disallow = tuple()
            if 'disallow' in options:
                disallow = options['disallow']
                del options['disallow']

            self.filters.append((self.get_module(name, 'get_filter')(options), disallow))

    def get_module(self, module, accessor):
        """Get module."""

        if isinstance(module, util.string_type):
            mod = importlib.import_module(module)
        attr = getattr(mod, accessor, None)
        if not attr:
            raise ValueError("Could not find accessor '%s'!" % accessor)
        return attr()

    def check(self):
        """Walk source and initiate spell check."""

        for documents in self.documents:
            if self.name and self.name != documents.get('name', ''):
                continue

            # Perform spell check
            self.log('\nSpell Checking %s...' % documents.get('name', ''), 1)

            # Setup parser and variables for the spell check
            parser = self.get_module(documents['parser'], 'get_parser')(
                documents.get('options', {}), documents.get('default_encoding', 'ascii')
            )
            self.file_patterns = documents.get('file_patterns', [])

            options = self.setup_spellchecker(documents)
            output = self.setup_dictionary(documents)
            self.get_filters(documents)
            self.setup_excludes(documents)

            for sources in self.walk_src(documents.get('sources', []), parser):
                for result in self.check_spelling(sources, options, output):
                    yield result


# class Hunspell(Aspell):
#     """Hunspell spell check class."""

#     def setup_spellchecker(self, documents):
#         """Setup spell checker."""

#         return documents.get('hunspell', {})

#     def setup_dictionary(self, documents):
#         """Setup dictionary."""

#         dictionary_options = documents.get('dictionary', {})
#         output = os.path.abspath(dictionary_options.get('output', self.dict_bin))
#         lang = dictionary_options.get('lang', 'en_US')
#         wordlists = dictionary_options.get('wordlists', [])
#         if lang and wordlists:
#             self.compile_dictionary(lang, dictionary_options.get('wordlists', []), output)
#         else:
#             output = None
#         return output

#     def compile_dictionary(self, lang, wordlists, output):
#         """Compile user dictionary."""

#         output_location = os.path.dirname(output)
#         if not os.path.exists(output_location):
#             os.makedirs(output_location)
#         if os.path.exists(output):
#             os.remove(output)

#         self.log("Compiling Dictionary...", 1)
#         # Read word lists and create a unique set of words
#         words = set()
#         for wordlist in wordlists:
#             with open(wordlist, 'rb') as src:
#                 for word in src.read().split(b'\n'):
#                     words.add(word.replace(b'\r', b''))

#         # Sort and create wordlist
#         with open(self.dict_bin, 'wb') as dest:
#             dest.write(b'\n'.join(sorted(words)) + b'\n')

#     def setup_command(self, encoding, options, personal_dict):
#         """Setup command."""

#         cmd = [
#             'hunspell',
#             '-l',
#             '-i', codecs.lookup(encoding).name
#         ]

#         if personal_dict:
#             cmd.extend(['-p', personal_dict])

#         allowed = {
#             'check-apostrophe', 'check-url',
#             'i', 'd' 'H', 'n', 'o', 'r', 't', 'X'
#         }

#         for k, v in options.items():
#             if k in allowed:
#                 key = ('-%s' if len(k) == 1 else '--%s') % k
#                 if isinstance(v, bool) and v is True:
#                     cmd.append(key)
#                 elif isinstance(v, util.ustr):
#                     cmd.extend([key, v])
#                 elif isinstance(v, int):
#                     cmd.extend([key, util.ustr(v)])
#                 elif isinstance(v, list):
#                     for value in v:
#                         cmd.extend([key, util.ustr(value)])
#         return cmd



"""Filters."""
from __future__ import unicode_literals


class Filter(object):
    """Base filter class."""

    def __init__(self, options):
        """Initialization."""

    def filter(self, text, encoding):
        """Execute filter."""

        return text



"""Markdown filter."""
from __future__ import unicode_literals
from .. import filters
import markdown
from .. import util


class MarkdownFilter(filters.Filter):
    """Markdown filter."""

    def __init__(self, options):
        """Initialize."""

        extensions = []
        extension_configs = {}
        for item in options.get('markdown_extensions', []):
            if isinstance(item, util.ustr):
                extensions.append(item)
            else:
                k, v = list(item.items())[0]
                extensions.append(k)
                if v is not None:
                    extension_configs[k] = v
        self.markdown = markdown.Markdown(extensions=extensions, extension_configs=extension_configs)

    def filter(self, text, encoding):
        """Apply Markdown filter."""

        self.markdown.reset()
        return self.markdown.convert(text)


def get_filter():
    """Return the filter."""

    return MarkdownFilter



"""HTML filter."""
from __future__ import unicode_literals
from .. import filters
from collections import namedtuple
import bs4
import re
from .. import util

if util.PY3:
    from html.parser import HTMLParser
else:
    from HTMLParser import HTMLParser

RE_SELECTOR = re.compile(r'''(\#|\.)?[-\w]+|\*|\[([\w\-:]+)(?:([~^|*$]?=)(\"[^"]+\"|'[^']'|[^'"\[\]]+))?\]''')


class Selector(namedtuple('IgnoreRule', ['tag', 'id', 'classes', 'attributes'])):
    """Ignore rule."""


class SelectorAttribute(namedtuple('AttrRule', ['attribute', 'pattern'])):
    """Selector attribute rule."""


class HTMLFilter(filters.Filter):
    """HTML filter."""

    def __init__(self, options):
        """Initialize."""

        self.html_parser = HTMLParser()
        self.comments = options.get('comments', True) is True
        self.attributes = set(options.get('attributes', []))
        self.selectors = self.process_selectors(*options.get('ignores', []))

    def process_selectors(self, *args):
        """
        Process selectors.

        We do our own selectors as BeautifulSoup4 has some annoying quirks,
        and we don't really need to do nth selectors or siblings or
        descendants etc.
        """

        selectors = [
            Selector('style', None, tuple(), tuple()),
            Selector('script', None, tuple(), tuple()),
        ]

        for selector in args:
            tag = None
            tag_id = None
            classes = set()
            attributes = []

            for m in RE_SELECTOR.finditer(selector):
                if m.group(2):
                    attr = m.group(2).lower()
                    op = m.group(3)
                    if op:
                        value = m.group(4)[1:-1] if m.group(4).startswith('"') else m.group(4)
                    else:
                        value = None
                    if not op:
                        # Attribute name
                        pattern = None
                    elif op.startswith('^'):
                        # Value start with
                        pattern = re.compile(r'^%s.*' % re.escape(value))
                    elif op.startswith('$'):
                        # Value ends with
                        pattern = re.compile(r'.*?%s$' % re.escape(value))
                    elif op.startswith('*'):
                        # Value contains
                        pattern = re.compile(r'.*?%s.*' % re.escape(value))
                    elif op.startswith('~'):
                        # Value contains word within space separated list
                        pattern = re.compile(r'.*?(?:(?<=^)|(?<= ))%s(?=(?:[ ]|$)).*' % re.escape(value))
                    elif op.startswith('|'):
                        # Value starts with word in dash separated list
                        pattern = re.compile(r'^%s(?=-).*' % re.escape(value))
                    else:
                        # Value matches
                        pattern = re.compile(r'^%s$' % re.escape(value))
                    attributes.append(SelectorAttribute(attr, pattern))
                else:
                    selector = m.group(0).lower()
                    if selector.startswith('.'):
                        classes.add(selector[1:].lower())
                    elif selector.startswith('#') and tag_id is None:
                        tag_id = selector[1:]
                    elif tag is None:
                        tag = selector
                    else:
                        raise ValueError('Bad selector!')

            if tag or tag_id or classes:
                selectors.append(Selector(tag, tag_id, tuple(classes), tuple(attributes)))

        return selectors

    def skip_tag(self, el):
        """Determine if tag should be skipped."""

        skip = False
        for selector in self.selectors:
            if selector.tag and selector.tag not in (el.name.lower(), '*'):
                continue
            if selector.id and selector.id != el.attrs.get('id', '').lower():
                continue
            if selector.classes:
                current_classes = [c.lower() for c in el.attrs.get('class', [])]
                found = True
                for c in selector.classes:
                    if c not in current_classes:
                        found = False
                        break
                if not found:
                    continue
            if selector.attributes:
                found = True
                for a in selector.attributes:
                    value = el.attrs.get(a.attribute)
                    if isinstance(value, list):
                        value = ' '.join(value)
                    if not value:
                        found = False
                        break
                    elif a.pattern is None:
                        continue
                    elif a.pattern.match(value) is None:
                        found = False
                        break
                if not found:
                    continue
            skip = True
            break
        return skip

    def html_to_text(self, tree):
        """
        Parse the HTML creating a buffer with each tags content.

        Skip any selectors specified and include attributes if specified.
        Ignored tags will not have their attributes scanned either.
        """

        text = []
        attributes = []
        root = tree.name == '[document]'

        # Handle comments
        if isinstance(tree, bs4.Comment):
            if self.comments:
                string = util.ustr(tree).strip()
                if string:
                    text.append(string)
        elif root or not self.skip_tag(tree):
            # Check attributes for normal tags
            if not root:
                for attr in self.attributes:
                    value = tree.attrs.get(attr, '').strip()
                    if value:
                        attributes.append(value)

            # Walk children
            for child in tree:
                if isinstance(child, bs4.element.Tag):
                    if child.contents:
                        t, a = (self.html_to_text(child))
                        text.extend(t)
                        attributes.extend(a)
                # Get content if not the root and not a comment (unless we want comments).
                elif not root and (not isinstance(child, bs4.Comment) or self.comments):
                    string = util.ustr(child).strip()
                    if string:
                        text.append(string)

        if root:
            return self.html_parser.unescape(' '.join([' '.join(text), ' '.join(attributes)]))
        else:
            return (text, attributes)

    def filter(self, text, encoding):
        """Filter the text."""

        return self.html_to_text(bs4.BeautifulSoup(text, "html5lib"))


def get_filter():
    """Return the filter."""

    return HTMLFilter



"""Markdown parsing."""
from __future__ import unicode_literals
from .. import parsers
import codecs
from ..filters import markdown_filter


class MarkdownParser(parsers.Parser):
    """Spelling Python."""

    FILE_PATTERNS = ('*.md', '*.markdown')

    def __init__(self, options, default_encoding='ascii'):
        """Initialization."""

        self.filter = markdown_filter.MarkdownFilter(options)
        super(MarkdownParser, self).__init__(options, default_encoding)

    def parse_file(self, source_file, encoding):
        """Parse Markdown file."""

        with codecs.open(source_file, 'r', encoding=encoding) as f:
            text = f.read()
        content = [parsers.SourceText(self.filter.filter(text, encoding), source_file, encoding, 'markdown')]

        return content


def get_parser():
    """Return the parser."""

    return MarkdownParser



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setup pymdown-extensions."""

from setuptools import setup, find_packages
import os
import imp

LONG_DESC = '''
PySpelling is a spell checker that wraps around Aspell.
You can check out the list of available extensions and learn more about them by `reading the docs`_.

.. _`reading the docs`: http://facelessuser.github.io/pyspelling/

Support
=======

Help and support is available here at the repositories `bug tracker`_.
Please read about `support and contributing`_ before creating issues.

.. _`bug tracker`: https://github.com/facelessuser/pyspelling/issues
.. _`support and contributing`: http://facelessuser.github.io/pyspelling/contributing/
'''


def get_version():
    """Get version and version_info without importing the entire module."""

    devstatus = {
        'alpha': '3 - Alpha',
        'beta': '4 - Beta',
        'candidate': '4 - Beta',
        'final': '5 - Production/Stable'
    }
    path = os.path.join(os.path.dirname(__file__), 'pyspelling')
    fp, pathname, desc = imp.find_module('__version__', [path])
    try:
        v = imp.load_module('__version__', fp, pathname, desc)
        return v.version, devstatus[v.version_info[3]]
    finally:
        fp.close()


def get_requirements():
    """Get the dependencies."""

    with open("requirements/project.txt") as f:
        requirements = []
        for line in f.readlines():
            line = line.strip()
            if line and not line.startswith('#'):
                requirements.append(line)
    return requirements


VER, DEVSTATUS = get_version()

entry_points = {
    'console_scripts': [
        'pyspelling=pyspelling.__main__:main'
    ]
}

setup(
    name='pyspelling',
    version=VER,
    keywords='spelling',
    description='Spell checker.',
    long_description=LONG_DESC,
    author='Isaac Muse',
    author_email='Isaac.Muse@gmail.com',
    url='https://github.com/facelessuser/pyspelling',
    packages=find_packages(exclude=['tools', 'tests']),
    entry_points=entry_points,
    install_requires=get_requirements(),
    license='MIT License',
    classifiers=[
        'Development Status :: %s' % DEVSTATUS,
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



"""
Python parsing.

Parse Python docstrings.
"""
from __future__ import unicode_literals
from .. import parsers
import re
import textwrap
import tokenize
import ast
from .. import util

if util.PY3:
    tokenizer = tokenize.tokenize
    PREV_DOC_TOKENS = (tokenize.INDENT, tokenize.DEDENT, tokenize.NEWLINE, tokenize.ENCODING)
else:
    tokenizer = tokenize.generate_tokens
    PREV_DOC_TOKENS = (tokenize.INDENT, tokenize.DEDENT, tokenize.NEWLINE)

RE_PY_ENCODE = re.compile(
    br'^[^\r\n]*?coding[:=]\s*([-\w.]+)|[^\r\n]*?\r?\n[^\r\n]*?coding[:=]\s*([-\w.]+)'
)
RE_NON_PRINTABLE_ASCII = re.compile(br"[^ -~]+")


class PythonDecoder(parsers.Decoder):
    """Detect Python encoding."""

    def header_check(self, content):
        """Special Python encoding check."""

        encode = None

        m = RE_PY_ENCODE.match(content)
        if m:
            if m.group(1):
                encode = m.group(1).decode('ascii')
            elif m.group(2):
                encode = m.group(2).decode('ascii')
        if encode is None:
            encode = 'ascii'
        return encode


class PythonParser(parsers.Parser):
    """Spelling Python."""

    FILE_PATTERNS = ('*.py', '*.pyw')
    DECODER = PythonDecoder

    MODULE = 0
    FUNCTION = 1
    CLASS = 2

    def __init__(self, options, default_encoding='ascii'):
        """Initialization."""

        self.comments = options.get('comments', True) is True
        self.docstrings = options.get('docstrings', True) is True
        self.strings = options.get('strings', False) is True
        self.bytes = options.get('bytes', False) is True
        self.group_comments = options.get('group_comments', False) is True
        super(PythonParser, self).__init__(options, default_encoding)

    def is_py2_unicode_literals(self, text, source_file):
        """Check if Python 2 Unicode literals is used."""

        uliterals = False
        root = ast.parse(text, source_file)

        for node in ast.iter_child_nodes(root):
            if isinstance(node, ast.ImportFrom) and node.module:
                module = node.module.split('.')
            else:
                continue

            if module[0] != '__future__':
                continue

            for n in node.names:
                if n.name and n.name.split('.')[0] == 'unicode_literals':
                    uliterals = True
                    break
        return uliterals

    def get_ascii(self, text):
        """Retrieve ASCII text from byte string."""

        return RE_NON_PRINTABLE_ASCII.sub(r' ', text).decode('ascii')

    def parse_docstrings(self, source_file, encoding):
        """Retrieve the Python docstrings."""

        docstrings = []
        comments = []
        strings = []
        prev_token_type = tokenize.NEWLINE
        indent = ''
        name = None
        stack = [(source_file, 0, self.MODULE)]
        uliterals = True

        with open(source_file, 'rb') as source:

            if not util.PY3:
                uliterals = self.is_py2_unicode_literals(source.read(), source_file)
                source.seek(0)

            for token in tokenizer(source.readline):
                token_type = token[0]
                value = token[1]
                line = util.ustr(token[2][0])
                line_num = token[2][0]

                if util.PY3 and token_type == tokenize.ENCODING:
                    # PY3 will tell us for sure what our encoding is
                    encoding = value

                value = token[1]
                line = util.ustr(token[2][0])

                # Track function and class ancestry
                if token_type == tokenize.NAME:
                    if value in ('def', 'class'):
                        name = value
                    elif name:
                        parent = stack[-1][2]
                        prefix = ''
                        if parent != self.MODULE:
                            prefix = '.' if parent == self.CLASS else ', '
                        if name == 'class':
                            stack.append(('%s%s' % (prefix, value), len(indent), self.CLASS))
                        elif name == 'def':
                            stack.append(('%s%s()' % (prefix, value), len(indent), self.FUNCTION))
                        name = None

                if token_type == tokenize.COMMENT and self.comments:
                    # Capture comments
                    if (
                        self.group_comments and
                        prev_token_type == tokenize.NL and
                        comments and (comments[-1][2] + 1) == line_num
                    ):
                        # Group multiple consecutive comments
                        comments[-1][0] += '\n' + value[1:]
                        comments[-1][2] = line_num
                    else:
                        if len(stack) > 1:
                            loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                        else:
                            loc = "%s(%s)" % (stack[0][0], line)
                        comments.append([value[1:], loc, line_num])
                if token_type == tokenize.STRING:
                    # Capture docstrings
                    # If previously we captured an INDENT or NEWLINE previously we probably have a docstring.
                    # NL seems to be a different thing.
                    if prev_token_type in PREV_DOC_TOKENS:
                        if self.docstrings:
                            value = value.strip()
                            if not util.PY3 and not value.startswith((b'u', b'b')):
                                value = (b'u' if uliterals else b'b') + value
                            string = textwrap.dedent(eval(value))

                            if not isinstance(string, util.ustr):
                                # Since docstrings should be readable and printable,
                                # if byte string assume 'ascii'.
                                string = string.decode('ascii')
                            loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                            docstrings.append(parsers.SourceText(string, loc, encoding, 'docstring'))
                    elif self.strings:
                        value = value.strip()
                        if not util.PY3 and not value.startswith((b'u', b'b')):
                            value = (b'u' if uliterals else b'b') + value
                        string = textwrap.dedent(eval(value))
                        if isinstance(string, util.ustr) or self.bytes:
                            string_type = 'string'
                            if not isinstance(string, util.ustr):
                                string = self.get_ascii(string)
                                string_type = 'bytes'
                            loc = "%s(%s): %s" % (stack[0][0], line, ''.join([crumb[0] for crumb in stack[1:]]))
                            strings.append(parsers.SourceText(string, loc, encoding, string_type))

                if token_type == tokenize.INDENT:
                    indent = value
                elif token_type == tokenize.DEDENT:
                    indent = indent[:-4]
                    if len(stack) > 1 and len(indent) <= stack[-1][1]:
                        stack.pop()

                prev_token_type = token_type

        final_comments = []
        for comment in comments:
            final_comments.append(parsers.SourceText(textwrap.dedent(comment[0]), comment[1], encoding, 'comment'))

        return docstrings + final_comments + strings

    def parse_file(self, source_file, encoding):
        """Parse Python file returning content."""

        return self.parse_docstrings(source_file, encoding)


def get_parser():
    """Return the parser."""

    return PythonParser



"""Utilities."""
from __future__ import unicode_literals
import subprocess
import sys
import string
import random
from collections import namedtuple

PY3 = sys.version_info >= (3, 0)

if PY3:
    string_type = str
    ustr = str
    bstr = bytes
else:
    string_type = basestring  # noqa
    ustr = unicode  # noqa
    bstr = str


def console(cmd, input_file=None, input_text=None):
    """Call with arguments."""

    returncode = None
    output = None

    if sys.platform.startswith('win'):
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        process = subprocess.Popen(
            cmd,
            startupinfo=startupinfo,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            shell=False
        )
    else:
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            shell=False
        )

    if input_file is not None:
        with open(input_file, 'rb') as f:
            process.stdin.write(f.read())
    if input_text is not None:
        process.stdin.write(input_text)
    output = process.communicate()
    returncode = process.returncode

    assert returncode == 0, "Runtime Error: %s" % (
        output[0].rstrip().decode('utf-8') if PY3 else output[0]
    )

    return output[0].decode('utf-8') if PY3 else output[0]


def random_name_gen(size=6):
    """Generate a random python attribute name."""

    return ''.join(
        [random.choice(string.ascii_uppercase)] +
        [random.choice(string.ascii_uppercase + string.digits) for i in range(size - 1)]
    ) if size > 0 else ''


class Results(namedtuple('Results', ['words', 'context', 'category', 'error'])):
    """Results."""

    def __new__(cls, words, context, category, error=None):
        """Allow defaults."""

        return super(Results, cls).__new__(cls, words, context, category, error)
